from .get_tmy_epw_file import get_tmy_epw_file
from .get_noaa_isd_lite_file import get_noaa_isd_lite_file
from .meteorology import Meteorology
from .analyze_noaa_isd_lite_file import analyze_noaa_isd_lite_file

import tempfile
import pandas as pd
import numpy as np
import os
import pkg_resources
from typing import Tuple
from calendar import isleap

from ._logging import _logger

# We buffer this path so that we don't create tons of temporary directories if the function is called many
# times, and so that calling it multiple times with the same WMO/year combination won't result in the same
# file being generated multiple times.
_tempdir_amy_epw = tempfile.mkdtemp()

def create_amy_epw_file(
        wmo_index:int,
        year:int,

        *,
        max_records_to_interpolate:int,
        max_records_to_impute:int,
        max_missing_amy_rows:int = None,
        amy_epw_dir:str = None,
        tmy_epw_dir:str = None,
        amy_dir:str = None,
        amy_files:Tuple[str, str] = None,
        allow_downloads:bool = False
) -> str:
    """
    Combine data from a Typical Meteorological Year (TMY) EPW file and Actual Meteorological Year (AMY)
    observed data to generate an AMY EPW file for a single calendar year at a given WMO.
    :param wmo_index: The WMO Index of the weather station for which the EPW file should be generated.
        Currently only weather stations in the United States are supported.
    :param year: The year for which the EPW should be generated
    :param amy_epw_dir: The directory into which the generated AMY EPW file should be written.
        If not defined, a temporary directory will be created
    :param tmy_epw_dir: The source directory for TMY EPW files. If a file for the requested WMO Index is
        already present, it will be used. Otherwise a TMY EPW file will be downloaded (see this package's
        get_tmy_epw_file() function for details). If no directory is given, the package's default
        directory (in data/tmy_epw_files/ in the package's directory) will be used, which will allow AMY
        files to be reused for future calls instead of downloading them repeatedly, which is quite time
        consuming.
    :param amy_dir: The source directory for AMY files. If a file for the requested WMO Index and year
        is already present, it will be used. Otherwise a TMY EPW file will be downloaded (see this package's
        get_noaa_isd_lite_file() function for details). If no directory is given, the package's default
        directory (in data/ in the package's directory) will be used, which will allow AMY files to be
        reused for future calls instead of downloading them repeatedly, which is quite time consuming.
    :param amy_files: Instead of specifying amy_dir and allowing this method to try to find the appropriate
        file, you can use this argument to specify the actual files that should be used. There should be
        two files - the first the AMY file for "year", and the second the AMY file for the subsequent year,
        which is required to support shifting the timezone from GMT to the timezone of the observed meteorology.
    :param max_records_to_interpolate: The maximum length of sequence for which linear interpolation will be
        used to replace missing values. See the documentation of _handle_missing_values() below for details.
    :param max_records_to_impute: The maximum length of sequence for which imputation will be used to replace
        missing values. See the documentation of _handle_missing_values() below for details.
    :param max_missing_amy_rows: The maximum total number of missing rows to permit in a year's AMY file.
    :param allow_downloads: If this is set to True, then any missing TMY or AMY files required to generate the
        requested AMY EPW file will be downloaded from publicly available online catalogs. Otherwise, those files
        being missing will result in an error being raised.
    :return: The absolute path of the generated AMY EPW file
    """

    if amy_dir is not None and amy_files is not None:
        raise Exception("It is not possible to specify both amy_dir and amy_files")

    if amy_epw_dir is None:
        global _tempdir_amy_epw
        amy_epw_dir = _tempdir_amy_epw
        _logger.debug(f"No amy_epw_dir was specified - generated AMY EPWs will be stored in {amy_epw_dir}")

    # Either amy_files is specified, in which case we use the specified paths, or amy_dir is specified,
    # in which case we will search that directory for AMY files, or neither is specified, in which case
    # we will fall back to a generated temporary directory.
    if amy_files is not None:
        for p in amy_files:
            if not os.path.exists(p):
                raise Exception(f'Path {p} does not exist')

        amy_file_path, amy_next_year_file_path = amy_files
    else:
        if amy_dir is None:
            amy_dir = pkg_resources.resource_filename("diyepw", "data/noaa_isd_lite_files")
            _logger.debug(f"No amy_dir was specified - downloaded AMY files will be stored in the default location at {amy_dir}")

        amy_file_path = get_noaa_isd_lite_file(wmo_index, year, output_dir=amy_dir, allow_downloads=allow_downloads)
        amy_next_year_file_path = get_noaa_isd_lite_file(wmo_index, year+1, output_dir=amy_dir, allow_downloads=allow_downloads)

    if max_missing_amy_rows is not None:
        amy_file_analysis = analyze_noaa_isd_lite_file(amy_file_path)
        if amy_file_analysis['total_rows_missing'] > max_missing_amy_rows:
            raise Exception(f"File is missing {amy_file_analysis['total_rows_missing']} rows, but maximum allowed is {max_missing_amy_rows}")

    # Read in the corresponding TMY3 EPW file.
    tmy_epw_file_path = get_tmy_epw_file(wmo_index, tmy_epw_dir, allow_downloads=allow_downloads)
    tmy = Meteorology.from_tmy3_file(tmy_epw_file_path)

    # If the year we are generating an AMY EPW for is a leap year, then we need to add the leap day to the TMY data.
    # We'll do that by adding the day with all empty values, then using the same routine we do to interpolate/impute
    # missing data in our AMY files to fill in the missing data.
    if isleap(year):
        _logger.debug(f"{year} is a leap year, using the interpolation strategy to populate TMY data for Feb. 29")
        for hour in range(1, 25):
            col_names = tmy.observations.columns.to_list()
            new_row_vals = [1982, 2, 29, hour, 0]
            new_row_vals.extend(np.repeat(np.nan, len(col_names) - len(new_row_vals)))
            new_row = pd.DataFrame([new_row_vals], columns=col_names)
            tmy.observations = tmy.observations.append(new_row)

        # We sort by month, day and hour. We do *not* sort by year, because the year column doesn't matter and because
        # it is in any case not consistent throughout a TMY data set
        tmy.observations = tmy.observations.sort_values(by=["month", "day", "hour"])

        #TODO: This is where I left off Thursday night. This call is changing the data types of the date fields into
        # floating point values, which breaks the EPW file
        _handle_missing_values(
            tmy.observations,
            max_to_interpolate=0, # We only want the imputation strategy to be used for the 24 missing hours
            max_to_impute=24,
            step=1,
            imputation_range=14 * 24, # Two weeks, in hours
            imputation_step=24,
            ignore_columns=["Flags"] # The TMY files we use seem to be missing data for this field entirely
        )

    amy_epw_file_name = f"{tmy.country}_{tmy.state}_{tmy.city}.{tmy.station_number}_AMY_{year}.epw"
    amy_epw_file_name = amy_epw_file_name.replace(" ", "-")
    amy_epw_file_path = os.path.join(amy_epw_dir, amy_epw_file_name)

    if os.path.exists(amy_epw_file_path):
        _logger.debug(f"File already exists at {amy_epw_file_path}, so a new one won't be generated.")
        return amy_epw_file_path

    # Read in the NOAA AMY file for the station for the requested year as well as the first 23 hours (sufficient
    # to handle the largest possible timezone shift) of the subsequent year - the subsequent year's data will be
    # used to populate the last hours of the year because of the time shift that we perform, which moves the first
    # hours of January 1 into the final hours of December 31.
    amy_df = pd.read_csv(amy_file_path, delim_whitespace=True, header=None)
    amy_next_year_df = pd.read_csv(amy_next_year_file_path, delim_whitespace=True, header=None, nrows=23)
    amy_df = pd.concat([amy_df, amy_next_year_df]).reset_index(drop=True)

    amy_df = _set_noaa_df_columns(amy_df)
    amy_df = _create_timestamp_index_for_noaa_df(amy_df)

    # Shift the timestamp (index) to match the time zone of the WMO station.
    amy_df = amy_df.shift(periods= tmy.timezone_gmt_offset, freq='H')

    # Remove time steps that aren't applicable to the year of interest
    amy_df = _map_noaa_df_to_year(amy_df, year)

    _handle_missing_values(
        amy_df,
        step=pd.Timedelta("1h"),
        max_to_interpolate=max_records_to_interpolate,
        max_to_impute=max_records_to_impute,
        imputation_range=pd.Timedelta("2w"),
        imputation_step=pd.Timedelta("1d"),
        missing_values=[np.nan, -9999.]
    )

    # Initialize new column for station pressure (not strictly necessary)
    amy_df['Station_Pressure'] = None

    # Convert sea level pressure in NOAA df to atmospheric station pressure in Pa.
    for index in amy_df.index:
        stp = _convert_sea_level_pressure_to_station_pressure(amy_df['Sea_Level_Pressure'][index], tmy.elevation)
        amy_df.loc[index, 'Station_Pressure'] = stp

    # Change observation values to the values taken from the AMY data
    tmy.set('year', year)
    tmy.set('Tdb', [i / 10 for i in amy_df['Air_Temperature']])  # Convert AMY value to degrees C
    tmy.set('Tdew', [i / 10 for i in amy_df['Dew_Point_Temperature']])  # Convert AMY value to degrees C
    tmy.set('Patm', amy_df['Station_Pressure'])
    tmy.set('Wdir', amy_df['Wind_Direction'])
    tmy.set('Wspeed', [i / 10 for i in amy_df['Wind_Speed']])  # Convert AMY value to m/sec

    # Check for violations of EPW file standards
    epw_rule_violations = tmy.validate_against_epw_rules()
    if len(epw_rule_violations) > 0:
        raise Exception("EPW validation failed:\n" + "\n".join(epw_rule_violations))

    # Write new EPW file if no validation errors were found.
    tmy.write_epw(amy_epw_file_path)

    return amy_epw_file_path

def _set_noaa_df_columns(df:pd.DataFrame) -> pd.DataFrame:
    """Add headings to a NOAA ISD Lite formatted dataframe, and Drop columns for observations
    that won't be used in populating the EPW files.
    """
    list_of_columns = ["Year", "Month", "Day", "Hour", "Air_Temperature",
                       "Dew_Point_Temperature", "Sea_Level_Pressure", "Wind_Direction",
                       "Wind_Speed", "Sky_Condition_Total_Coverage_Code",
                       "Liquid_Precipitation_Depth_Dimension_1H", "Liquid_Precipitation_Depth_Dimension_6H"]
    df.columns = list_of_columns

    # Remove unnecessary columns
    df = df.drop(columns=[
        'Sky_Condition_Total_Coverage_Code',
        'Liquid_Precipitation_Depth_Dimension_1H',
        'Liquid_Precipitation_Depth_Dimension_6H'
    ])

    return df

def _create_timestamp_index_for_noaa_df(df:pd.DataFrame) -> pd.DataFrame:
    """Convert the year, month, day fields of a NOAA ISD Lite DataFrame into
    a timestamp and make that timestamp the index of the DataFrame
    :param df:
    :return:
    """
    df['timestamp'] = pd.to_datetime(pd.DataFrame({'year': df['Year'],
                                                   'month': df['Month'],
                                                   'day': df['Day'],
                                                   'hour': df['Hour']}))
    df = df.set_index('timestamp')

    # Remove unnecessary columns
    df = df.drop(columns=['Year', 'Month', 'Hour', 'Day'])

    return df

def _map_noaa_df_to_year(df, year):
    """Add headings to a NOAA ISD Lite formatted dataframe, convert year-month-day-hour columns to a timestamp,
    set the timestamp as index, and make sure each hour of the DF's range has a timestamp, regardless of whether
    there are any observations in that hour. Drop columns for observations that won't be used in populating the
    EPW files.

    The assumption of this function is that the dataframe ranges from the beginning of the year to some
    """
    # Create series of continuous timestamp values for that year
    all_timestamps = pd.date_range(str(year) + '-01-01 00:00:00', str(year) + '-12-31 23:00:00', freq='H')
    all_timestamps = pd.DataFrame(all_timestamps, columns=['timestamp'])

    # Merge to one dataframe containing all continuous timestamp values.
    df = pd.merge(all_timestamps, df, how='left', left_on='timestamp', right_index=True)
    df = df.set_index('timestamp')

    return df

def _handle_missing_values(
        df:pd.DataFrame, *, step, max_to_interpolate:int, max_to_impute:int,
        imputation_range, imputation_step, missing_values:list=None, ignore_columns=[]
):
    """
    Look for missing values in a DataFrame. If possible, the missing values will be
    populated in place, using one of two strategies:

    If the missing values are in a contiguous block up to the length defined by max_to_interpolate,
    the values will linearly interpolated using the previous and following values.

    Otherwise, if the missing values are in a contiguous block up to the length defined by
    max_to_impute, the values will be imputed by going back through the indices by
    imputation_range, then stepping through by step sizes defined by imputation_step
    until the index that is imputation_range ahead of the missing value is found, and
    averaging all values encountered. For example, assuming a dataframe indexed by timestamp,
    if imputation_range is two weeks and imputation_step is 24 hours, a missing value will
    be imputed by calculating the average value at the same time of day every day going back
    two weeks and forward two weeks from the missing row.

    Otherwise, if the DataFrame contains at least one contiguous block of missing values
    larger than max_to_impute, it will be left unchanged, and an Exception will be raised.

    :param df: The dataframe to be searched for missing values.
    :param step: The step size to use in considering whether the indexes of the dataframe are
      contiguous. If two indices are one step apart, they are neighbors in a contiguous block.
      Otherwise they do not belong to the same contiguous block.
    :param max_to_interpolate: The maximum length of contiguous block to treat with the
      interpolation strategy described above.
    :param max_to_impute: The maximum length of contiguous block to treat with the imputation
      strategy described above.
    :param imputation_range: The distance before and after a missing record that will be searched
      for values to average when imputing a missing value
    :param imputation_step: The step-size to use in finding values to impute from, as described
      in the imputation strategy above.
    :param missing_values: Values matching any value in this list will be treated as missing. If not
      passed, defaults to numpy.nan
    :param ignore_columns: Columns listed here will be skipped; their data will be left unchanged, and
      no checks will be made of whether they are missing overly large segments of values.
    :return:
    """

    if missing_values is None: # pragma: no cover - We don't currently have any calls that don't specify this argument
        missing_values = [np.nan]

    def get_indices_to_replace(df, col_name):
        indices_to_replace = df.index[df[col_name].isna()].tolist()
        indices_to_replace = _split_list_into_contiguous_segments(
            indices_to_replace,
            step=step
        )
        return indices_to_replace

    for col_name in (c for c in df if c not in ignore_columns):

        # TODO: This is the line that is breaking the dates. It seems like we dodged the bullet here for this long
        #   just because we always converted the date fields into an index before we added interpolating for the
        #   leap day data
        if df[col_name].isin(missing_values).any():
            df.loc[df[col_name].isin(missing_values), col_name] = np.nan

        indices_to_replace = get_indices_to_replace(df, col_name)

        # There is no work to be done on this column if it has no missing data
        if len(indices_to_replace) == 0: # pragma: no cover
            continue

        # max(..., key=len) gives us the longest sequence, then we use len() to get that sequence's length
        max_sequence_length = len(max(indices_to_replace, key=len))

        # We raise an exception if a column has too many sequential missing rows; it's up to the calling
        # code to decide how we are going to handle records that can't be processed for this reason.
        if max_sequence_length > max_to_impute: # pragma: no cover - We check for this case before this function is called
            raise Exception("The longest set of missing records for {} is {}, but the max allowed is {}".format(
                col_name, max_sequence_length, max_to_impute
            ))

        # We make two passes to fill in missing records: The first pass uses the imputation strategy described
        # in this function's doc comment to fill in any gaps that are larger than max_to_interpolate. That
        # pass leaves behind any sequences that are smaller than that limit, and also leaves behind the first
        # and last item in any imputed sequence, which are also interpolated (i.e. set to the average of the imputed
        # value and the observed value on either side) to smooth out the transition between computed and observed
        # values.
        for indices in indices_to_replace:
            # Any blocks within our interpolation limit are skipped - they'll be filled in by the interpolate()
            # call below
            if len(indices) <= max_to_interpolate:
                continue

            # We will perform imputation on all the elements in the chunk *except* for the first and last
            # ones, which will be interpolated to smooth out the transition between computed and observed values
            indices_to_impute = indices[1:-1]

            # Set each missing value to the average of all the values in the range extending from imputation_range
            # indices behind to imputation_range indices ahead, walking through that range in steps whose size are
            # set by imputation_step.
            for index_to_impute in indices_to_impute:
                replacement_value_index = index_to_impute - imputation_range
                replacement_values = []
                while replacement_value_index <= index_to_impute + imputation_range:
                    if replacement_value_index in df.index:
                        replacement_values.append(df[col_name][replacement_value_index])
                    replacement_value_index += imputation_step

                # Take the mean of the values pulled. Will ignore NaNs.
                df[col_name][index_to_impute] = pd.Series(replacement_values, dtype=np.float64).mean()

        # Perform interpolation on any remaining missing values. At this point we know that there are no
        # sequences larger than the max permitted for interpolation, because they would have been imputed
        # or caused an exception (if larger than the imputation limit), so we can just call interpolate()
        # on anything that is still missing.
        df[col_name].interpolate(inplace=True)

def _split_list_into_contiguous_segments(l:list, step):
    """
    Given a list, will return a new list of lists, where each of the inner lists is one block of contiguous
    values from the original list.

    Example: split_list_into_contiguous_segments([1, 2, 5, 6, 7, 9, 11], 1) =>
    [
        [1, 2],
        [5, 6, 7],
        [9],
        [11]
    ]

    :param l: The list to split. The values in this list must be of the same type as step_size, and must
      be of a type allowing sorting, as well as addition, such that you would expect some list item added to
      step_size to produce another valid list value. If the list contains duplicate entries, the duplicates
      will be removed.
    :param step: Items in the list that differ from one another by this amount will be considered
      neighbors in a contiguous segment.
    :return:
    """

    # Ensure the list is sorted and remove any duplicates
    l = list(set(l))
    l.sort()

    segments = []
    cur_segment = []
    prev_val = None
    for val in l:
        if prev_val is not None and val - step != prev_val:
            segments.append(cur_segment)
            cur_segment = [val]
        else:
            cur_segment.append(val)
        prev_val = val
    if len(cur_segment) > 0:
        segments.append(cur_segment)

    return segments

def _convert_sea_level_pressure_to_station_pressure(Pa, h_m) -> object:
    """Return the atmospheric station pressure in Pa given sea level pressure in hPa*10 and station elevation in m."""

    # convert (or keep) pressure and elevation inputs as floats
    Pa = float(Pa)
    h_m = float(h_m)

    # convert from hectopascals*10 to inHg
    Pa_inHg = Pa/10 * 0.029529983071445

    # calculate station pressure according to formula from https://www.weather.gov/epz/wxcalc_stationpressure
    Pstn_inHg = Pa_inHg * ((288 - 0.0065*h_m)/288)**5.2561

    # convert from inHg to Pa
    Pstn = Pstn_inHg * 3386.389

    return Pstn


import unittest
import diyepw
import pkg_resources
import tempfile
import os


class TmyGetNoaaIsdLiteFileTest(unittest.TestCase):
    """
    Tests of the behavior of the get_noaa_isd_lite_file() function
    """

    def test_file_download(self):
        """
        Confirm the function works correctly if all files (the catalog and the file itself) must be downloaded.
        """
        with tempfile.TemporaryDirectory() as output_dir:
            wmo_index = 722016
            year = 2015

            # If we don't pass the allow_downloads flag to give explicit permission, the function should recognize that
            # the requested file is not present in output_dir (which is empty) and raise an Exception rather than proceed
            with self.assertRaises(Exception):
                diyepw.get_noaa_isd_lite_file(wmo_index, year, output_dir=output_dir)

            # Now pass the allow_downloads flag and check that the download behavior works correctly.
            # Call the method twice to exercise the shortcut in the code that will skip the download if
            # the file is already present in output_dir. This is difficult to confirm (unless we wanted to verify
            # that the second call ran faster?) but can be verified by seeing in the coverage report that the
            # code branch associated with that behavior has been invoked.
            for _ in range(2):
                file = diyepw.get_noaa_isd_lite_file(wmo_index, year, output_dir=output_dir, allow_downloads=True)
                self._validate_isd_lite_file(file)

        # We need a new temp directory because we need to trigger another download attempt
        with tempfile.TemporaryDirectory() as output_dir:
            # Delete the catalog data file, so that the next call to get_tmy_epw_file() will be forced to trigger
            # the file to be downloaded anew
            os.unlink(pkg_resources.resource_filename("diyepw", f"data/noaa_isd_lite_catalogs/{year}"))

            # With the catalog missing and a fresh temporary output directory, we get the same behavior as above
            # when trying to get a TMY EPW file without passing allow_downloads, but in this case it is due to
            # the catalog failing to download.
            with self.assertRaises(Exception):
                diyepw.get_noaa_isd_lite_file(wmo_index, year, output_dir=output_dir)

            # Now call it with allow_downloads to ensure that the catalog downloads correctly. Here we
            # don't have to do two calls as we did above because the "catalog already exists" shortcut
            # will have been exercised by at least one of the two calls in the above loop, if not both.
            file = diyepw.get_noaa_isd_lite_file(wmo_index, year, output_dir=output_dir, allow_downloads=True)
            self._validate_isd_lite_file(file)

    def _validate_isd_lite_file(self, file_path:str):
        """Perform some assertions on a file to verify that it is a valid ISD Lite file"""

        # analyze_noaa_isd_lite_file() interacts with the contents of a file deeply enough
        # to serve as a good approximation for validating that the file is well-formatted,
        # since the function will raise an Exception if it's unable to read data from the
        # file due to it being incorrectly formatted.
        analysis = diyepw.analyze_noaa_isd_lite_file(file_path)
        for field_name in ['file', 'total_rows_missing', 'max_consec_rows_missing']:
            self.assertIn(field_name, analysis)
        self.assertEqual(analysis['file'], file_path)

if __name__ == '__main__': # pragma: no cover
    unittest.main()



import unittest
import diyepw
import random
import tempfile
import pvlib
import pandas as pd
import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

class MeteorologyTest(unittest.TestCase):
    """
    Tests of the behavior of the diyepw.Meteorology class
    """

    def setUp(self) -> None:
        super().setUp()
        tmy_file_path = os.path.join(THIS_DIR, 'files', 'TEST_TMY3.epw')
        self._meteorology = diyepw.Meteorology.from_tmy3_file(tmy_file_path)

    def test_observation_setter(self):
        # We should get an exception if we use set() to set a non-existent observation column
        with self.assertRaises(Exception):
            self._meteorology.set("not_a_real_column", -1)

        # We should get an exception if we use set() to set an observation column to an array
        # with a number of values differing from the number of observation rows already present
        with self.assertRaises(Exception):
            self._meteorology.set('Tdb', [1, 2, 3])

    def test_write_epw(self):
        """
        Confirm that an EPW file generated by the Meteorology.write_epw() method is a valid EPW file
        :return:
        """
        with tempfile.NamedTemporaryFile('w+') as tmp_file:
            self._meteorology.write_epw(tmp_file.name)
            with open(tmp_file.name, "r") as m:
                try:
                    parsed_epw, col_names = pvlib.iotools.parse_epw(m)
                except Exception as e:
                    raise Exception(f"Encountered an error trying to parse the produce of Meteorology.write_epw() as an EPW file: {e}")

                # Make sure that parse_epw() actually succeeded in creating a DataFrame instance
                self.assertIsInstance(parsed_epw, pd.DataFrame)


    def test_property_setters_and_getters__good_values(self):
        """
        Tests the Meteorology class's @property-tagged setter/getter methods, confirming that
        when valid values are submitted, they properly store the value and the same value is returned
        from the equivalent getter method
        :return:
        """
        test_values = {
            'station_number' : [123456, 654321],
            'latlong' : [(85.2, -17.3), (-12.4, 18)],
            'city' : ['Richland', 'Neuenhaus'],
            'state' : ['WA', 'AL'],
            'country' : ["United States of America", "Peoples' Republic of Testcaseistan"],
            'timezone_gmt_offset' : [-6, 8],
            'elevation' : [2431, -5]
        }

        # Assert that, for each test value defined above, the identical value is returned from the
        # property getter after being assigned by the property setter
        for property_name in test_values:
            for test_value in test_values[property_name]:
                setattr(self._meteorology, property_name, test_value)
                self.assertEqual(test_value, getattr(self._meteorology, property_name))

    def test_property_setters__bad_values(self):
        """
        Tests that the Meteorology class's @property-tagged setter methods properly raise Exceptions when
        passed invalid values
        :return:
        """
        test_values = {
            'station_number' : [
                -6, # Negative
                "Banana", # Not numerical
                44321, # Too short for a WMO index
                8888888 # Too long for a WMO index
            ],
            'latlong' : [5, (91, 0), (0, -181)], # Latlong has to be a duple of valid lat/long values
            'timezone_gmt_offset' : [-13, 13], # Timezone offsets must be in the range +/-12
            'elevation' : ["a frippery"] # Elevation must be an integer
        }

        # Assert that, for each test value defined above, an Exception is raised by the attempt to set it
        for property_name in test_values:
            for test_value in test_values[property_name]:
                with self.assertRaises(
                        Exception,
                        msg=f"Setting Meteorology property {property_name} to {test_value} d"
                            f"id not cause an Exception to be raised"
                ):
                    setattr(self._meteorology, property_name, test_value)

    def test_create_from_tmy3(self):
        """
        Tests that observations from a known TMY3 file result in an instance with the expected values
        :return:
        """
        self.assertEqual(self._meteorology.city, "Testville")
        self.assertEqual(self._meteorology.country, "USA")
        self.assertEqual(self._meteorology.elevation, 78.)
        self.assertEqual(self._meteorology.latlong, (32.1, -90.23009))
        self.assertEqual(self._meteorology.station_number, 799999)
        self.assertEqual(self._meteorology.timezone_gmt_offset, -6)

        observations = self._meteorology.observations
        expected_columns = [
            'year', 'month', 'day', 'hour', 'minute', 'Tdb', 'Tdew', 'RH', 'Patm', 'ExHorRad', 'ExDirNormRad', 'HorIR',
            'GHRad', 'DNRad', 'DHRad', 'GHIll', 'DNIll', 'DHIll', 'ZenLum', 'Wdir', 'Wspeed', 'TotSkyCover',
            'OpSkyCover', 'Visib', 'CeilH', 'PresWeathObs', 'PresWeathCodes', 'PrecWater', 'AerOptDepth', 'SnowDepth',
            'DSLS', 'Albedo', 'LiqPrecDepth', 'LiqPrecQuant'
        ]
        for col in expected_columns:
            self.assertIn(col, observations.columns)

    def test_validation(self):
        """
        Tests that the validation rules are properly applied
        :return:
        """
        # Initially there should be no validation errors in the test file
        epw_violations = self._meteorology.validate_against_epw_rules()
        self.assertEqual(len(epw_violations), 0, msg=f"Expected 0 EPW validation errors but got {len(epw_violations)}")

        invalid_values = {
            'Tdb': [-71, 71],
            'Tdew': [-71, 71],
            'Patm': [30999, 120001],
            'Wspeed': [-1, 41],
            'Wdir': [-1, 361]
        }
        # Intentionally introduce validation errors and confirm that the expected error appears
        for col in invalid_values:
            original_values = self._meteorology.observations.loc[:, col].copy()
            for value in invalid_values[col]:
                changed_values = original_values.copy()
                changed_values.iloc[random.randint(0, len(changed_values) - 1)] = value
                self._meteorology.set(col, changed_values)
                epw_violations = self._meteorology.validate_against_epw_rules()
                self.assertEqual(
                    len(epw_violations),
                    1,
                    f"Expected 1 EPW validation error but got {len(epw_violations)}. Violations were: {epw_violations}"
                )
                self.assertIn(f"{col} must be in the range", epw_violations[0])

                # Replace the original values after each test so that only a single error is ever present
                self._meteorology.set(col, original_values)


if __name__ == '__main__': # pragma: no cover
    unittest.main()



import tempfile
import urllib.request as request
import re
import os
import shutil
import pkg_resources
import pandas as pd
from zipfile import ZipFile
from urllib.error import URLError

from ._logging import _logger
from .exceptions import DownloadNotAllowedError

def get_tmy_epw_file(wmo_index:int, output_dir:str = None, allow_downloads:bool = False):
    """
    Given a WMO index, retrieve a TMY (typical meteorological year) EPW file for that location
    :param wmo_index:
    :param output_dir: Optional output directory - if not specified, the file will be saved to a directory in the
        diyepw package so that it can be reused in the future if needed. If the directory already contains a TMY
        EWP file matching the requested WMO Index, then a new file will not be downloaded, we will just return
        that file's path
    :param allow_downloads: Set to True to allow downloads if TMY3 files or the catalog containing their
        names and URLs to be downloaded from http://climate.onebuilding.org if not already present. If you want
        to use this method without permitting downloads, then the TMY3 files can be downloaded manually and put
        into the directory specified by output_dir (which defaults to data/tmy_epw_files/ in the package's own
        directory structure). In this case, the files must have names identical to those used by
        http://climate.onebuilding.org/WMO_Region_4_North_and_Central_America/USA_United_States_of_America/
    :return: The path to the TMY EPW file
    """
    _logger.debug(f"get_tmy_epw_file() - Retrieving TMY EPW file for WMO {wmo_index}")

    if output_dir is None: # pragma: no cover - Not worth the risk of having tests that fiddle with real project data
        output_dir = pkg_resources.resource_filename("diyepw", "data/tmy_epw_files")

    if not os.path.isdir(output_dir):
        raise Exception(f'The path {output_dir} is not a valid directory path')

    tmy3_file_catalog = _get_tmy3_file_catalog(allow_downloads=allow_downloads)
    tmy3_file_information = tmy3_file_catalog.loc[tmy3_file_catalog['wmo_index'] == int(wmo_index)]

    if len(tmy3_file_information) == 0:
        raise Exception(f"No TMY3 file appears to exist in our catalog for the WMO Index '{wmo_index}'")

    # Get the filename for the requested WMO index's TMY3 file. We swap the extension
    epw_file_name = tmy3_file_information.iloc[0]['file_name']
    epw_file_url  = tmy3_file_information.iloc[0]['url']
    epw_file_path = os.path.join(output_dir, epw_file_name)

    # Check whether we already have the TMY EPW file in our directory. If we do it would be a waste of time
    # to download it again.
    if os.path.exists(epw_file_path):
        _logger.debug(f"TMY EPW file ({epw_file_path}) already exists, won't download again.")
        return epw_file_path

    # If the TMY EPW file is not already present in our directory, then we need to check that we've been given
    # permission to download missing files
    if not allow_downloads:
        raise DownloadNotAllowedError(
            f"The TMY3 file {epw_file_path} is not present. Pass allow_downloads=True to "
            f"allow the missing data to be automatically downloaded from {epw_file_url}"
        )

    # Download the ZIP file and decompress it. It contains a number of files including the EPW that we are looking for.
    tmp_file_handle, tmp_file_path = tempfile.mkstemp()
    tmp_dir = tempfile.mkdtemp()
    _logger.debug(f"Downloading file from {epw_file_url} and saving to {epw_file_path}")
    try:
        with request.urlopen(epw_file_url) as response:
            with open(tmp_file_handle, 'wb') as downloaded_file:
                downloaded_file.write(response.read())
    except URLError: # pragma: no cover - Not worth the time to write a test that disables the internet just for one exception
        raise Exception(f"Failed to download TMY EPW file {epw_file_url} - are you connected to the internet?")
    except Exception as e: # pragma: no cover - This block is good for adding information to an Exception but can't be intentionally provoked
        raise Exception(f"Error downloading from {epw_file_url}: {e}")

    with ZipFile(tmp_file_path, 'r') as zip_file:
        zip_file.extractall(tmp_dir)

    # Move the EPW file from the temporary directory into which we extracted
    # the ZIP file into the directory storing our EPWs
    os.rename(os.path.join(tmp_dir, epw_file_name), epw_file_path)

    # Delete the temporary files created in this call
    os.unlink(tmp_file_path)
    shutil.rmtree(tmp_dir)

    return epw_file_path

def _get_tmy3_file_catalog(allow_downloads:bool = False) -> pd.DataFrame:
    """
    Retrieves the catalog of TMY3 files, which contains a list of all WMOs in the United States, along with the
    names and download URLs of a TMY3 file for that WMO.

    :param allow_downloads: If True, and if the catalog is missing, then the catalog will be downloaded from
    http://climate.onebuilding.org and stored in this package's data directory. The catalog should not actually
    be missing unless it has been manually removed, as the DIYEPW package ships with a copy of the file already
    included.

    :return: A pandas DataFrame with the following columns:
        - wmo_index: The WMO Index of a weather station
        - file_name: The name of the TMY3 file corresponding to that WMO Index
        - url: The URL from which the TMY3 file can be downloaded
    """

    catalog_file_path = pkg_resources.resource_filename("diyepw", "data/tmy_epw_catalogs/tmy_epw_catalog.csv")

    # The sources we know of for TMY EPW files are http://climate.onebuilding.org and https://energyplus.net/weather;
    # we use the climate.onebuilding.org source here because it has all of the EPW files linked
    # from a single page, making it relatively easy to work with
    catalog_url = "http://climate.onebuilding.org/WMO_Region_4_North_and_Central_America/USA_United_States_of_America/"

    if os.path.exists(catalog_file_path):
        catalog = pd.read_csv(catalog_file_path)
    else:
        # If the catalog is not already present, and the caller has allowed for downloads, download the catalog file
        # from the source site

        if not allow_downloads:
            raise DownloadNotAllowedError(
                f"The TMY3 catalog file {catalog_file_path} is not present. Pass allow_downloads=True to "
                f"allow the missing data to be automatically downloaded from {catalog_url}"
            )

        _logger.info(f"The TMY3 catalog file was not found at {catalog_file_path}, and allow_downloads is True, so "
                     f"the catalog will be downloaded from {catalog_url}")

        # Retrieve the TMY EPW catalog for the requested year.
        try:
            with request.urlopen(catalog_url) as response:
                catalog_html = response.read().decode('utf-8')
        except URLError: # pragma: no cover - Not worth the time to write a test that disables the internet just for one exception
            raise Exception(f"Failed to connect to {catalog_url} - are you connected to the internet?")
        except Exception as e: # pragma: no cover - This block is good for adding information to an Exception but can't be intentionally provoked
            raise Exception(f"Error downloading from {catalog_url}: {e}")

        # Iterate over each line in the catalog HTML page, parsing out the file names for each TMY3 file that is
        # encountered
        catalog = pd.DataFrame(columns=['wmo_index', 'file_name', 'url'])
        for line in catalog_html.splitlines():
            # Regex: Match hrefs pointing to files in the form *.#_TMY3.zip, where the # is the WMO
            # represented by the file.
            # Capture groups: The big capture group gets the file name, and the small one gets the WMO
            for match in re.finditer(r'href="([^"]*\.(\d{6})_TMY3.zip)"', line):
                file_path, wmo_index = match.groups()

                # The file extension of the download URL is .zip, and points to an archive file containing a number
                # of files. Within that archive is the TMY EPW file, which has the same name and the .epw extension.
                file_name = file_path.split('/')[-1].replace('.zip', '.epw')

                catalog = catalog.append({
                    'wmo_index': int(wmo_index),
                    'file_name': file_name,
                    'url': catalog_url + '/' + file_path
                }, ignore_index=True)

        catalog.to_csv(catalog_file_path, index=False)

    return catalog



import urllib.request as request
import re
import os
import pandas as pd
import pkg_resources
from urllib.error import URLError

from ._logging import _logger
from .exceptions import DownloadNotAllowedError

def get_noaa_isd_lite_file(wmo_index:int, year:int, *, output_dir:str = None, allow_downloads:bool = False) -> str:
    """
    Given a WMO index and a year, retrieve the corresponding NOAA ISD Lite AMY file
    :param wmo_index:
    :param year:
    :param output_dir: Optional output directory - if not specified, the file will be saved to a package directory.
        If the directory already contains a NOAA ISD Lite file matching the requested WMO Index and year, then a new
        file will not be downloaded from NOAA and that file's path will be returned
    :param allow_downloads: Pass True to permit NOAA ISD Lite files and related information to be downloaded from
        ncdc.noaa.gov if they are not already present in output_dir.
    :return: The path to the NOAA ISD Lite file
    """
    if output_dir is None: # pragma: no cover
        output_dir = pkg_resources.resource_filename('diyepw', 'data/noaa_isd_lite_files')
        _logger.debug(f"get_noaa_isd_lite_file() - output_dir was not defined, will use {output_dir}")

    if not os.path.exists(output_dir): # pragma: no cover
        os.mkdir(output_dir)
        _logger.debug(f"get_noaa_isd_lite_file() - {output_dir} did not exist, so has been created")

    # On the NOAA website, the ISD Lite files are named with a third number between WMO and year, but
    # since we don't use that third number for anything and it complicates identifying a file for a
    # WMO/Year combination, we simplify the name to only contain the values we care about
    file_name = f"{wmo_index}-{year}.gz"
    file_path = os.path.join(output_dir, file_name)

    # Download the ISD Lite file if it's not already in the output directory
    if not os.path.exists(file_path):
        url = _get_noaa_isd_lite_file_url(year, wmo_index, allow_downloads)

        if not allow_downloads:
            raise DownloadNotAllowedError(
                f"The ISD Lite file {file_path} is not present. Pass allow_downloads=True to allow the "
                f"missing data to be automatically downloaded from {url}"
            )

        try:
            with request.urlopen(url) as response:
                with open(file_path, 'wb') as downloaded_file:
                    downloaded_file.write(response.read())
        except URLError as e: # pragma: no cover
            raise Exception(f'Failed to download {url} - are you connected to the internet?')
        except Exception as e: # pragma: no cover
            raise Exception(f"Error downloading from {url}: {e}")

    return file_path

def _get_noaa_isd_lite_file_url(year:int, wmo_index:int, allow_downloads:bool) -> str:
    catalog = _get_noaa_isd_lite_file_catalog(year, allow_downloads=allow_downloads)
    wmo_index_row = catalog.loc[catalog['wmo_index'] == wmo_index]

    if len(wmo_index_row) == 0:
        raise Exception(f"Invalid WMO Inex: The NOAA ISD Lite catalog does not contain an entry for WMO Index {wmo_index}")

    file_name = wmo_index_row['file_name'].iloc[0]
    return f"https://www1.ncdc.noaa.gov/pub/data/noaa/isd-lite/{year}/{file_name}"

def _get_noaa_isd_lite_file_catalog(year:int, *, catalog_dir=None, allow_downloads:bool = False) -> pd.DataFrame:
    """
    Retrieve the list of all NOAA ISD Lite files for North America (WMO indices starting with 7) for a given year.
    If the file is not already present, one will be downloaded. Files are named after the year whose files they
    describe.
    :param year:
    :param catalog_dir: The directory in which to look for the file, and into which the file will be written if
        downloaded
    :param allow_downloads: Pass True to permit the catalog of available NOAA ISD Lite files for North America to
        be downloaded if it is not already present in catalog_dir
    :return: A Pandas Dataframe containing a set of file names. The file names can be
        appended to the URL https://www1.ncdc.noaa.gov/pub/data/noaa/isd-lite/{year}/ to download the files from
        NOAA
    """
    if catalog_dir is None:
        catalog_dir = pkg_resources.resource_filename('diyepw', 'data/noaa_isd_lite_catalogs')
        _logger.debug(f"catalog_dir was not defined, using {catalog_dir}")

    if not os.path.exists(catalog_dir): # pragma: no cover
        raise Exception(f"Directory {catalog_dir} does not exist")

    file_path = os.path.join(catalog_dir, str(year))

    # If the catalog file already exists, we'll read it. If it doesn't, we'll download it, import it into a
    # dataframe, and then save that so that it exists the next time we need it.
    if os.path.exists(file_path):
        _logger.debug(f"Catalog file exists at {file_path}, using it instead of downloading it from NOAA")
        catalog = pd.read_csv(file_path)
    else:
        catalog_url = f"https://www1.ncdc.noaa.gov/pub/data/noaa/isd-lite/{year}/"

        if not allow_downloads:
            raise DownloadNotAllowedError(
                f"The ISD Lite catalog file {file_path} is not present. Pass allow_downloads=True "
                f"to allow the missing data to be automatically downloaded from {catalog_url}"
            )

        _logger.info(f"Downloading catalog file for year {year} from {catalog_url}")

        # Retrieve the NOAA ISD Lite catalog for the requested year
        catalog = pd.DataFrame(columns=['wmo_index', 'file_name'])
        with request.urlopen(f"https://www1.ncdc.noaa.gov/pub/data/noaa/isd-lite/{year}/") as response:
            # Process the file: Look for an href linking to a file that starts with "7" (indicating it is
            # a North American WMO) and put all such referenced file names into the catalog file
            html = response.read().decode('utf-8')
            for line in html.splitlines():
                # Regex: Match hrefs pointing to files in the form #-#-#.gz, where the first # starts with a 7.
                # Capture groups: The big capture group gets the file name, and the small one gets the WMO
                for match in re.finditer(f'href="((7\d+)-.*\.gz)"', line):
                    file_name, wmo_index = match.groups()
                    catalog = catalog.append({'wmo_index': int(wmo_index), 'file_name': file_name}, ignore_index=True)

            catalog.to_csv(file_path, index=False)
            _logger.info(f"Catalog file for year {year} saved to {file_path}")

    return catalog



import unittest
import diyepw
import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class AnalyzeNoaaIsdLiteFileTest(unittest.TestCase):
    """
    Tests of the behavior of the analyze_noaa_isd_lite_file function
    """

    def test_file_without_missing_rows(self):
        """
        Confirm the correct behavior of the script when handling an ISD Lite file that has no missing rows
        """
        isd_lite_file = os.path.join(THIS_DIR, 'files', 'noaa_isd_lite', 'uncompressed_no_missing_rows')

        analysis = diyepw.analyze_noaa_isd_lite_file(isd_lite_file)
        self._assert_analysis_for_no_missing_rows(analysis, isd_lite_file)

    def test_compression_methods(self):
        """
        Confirm that the "compression" argument can be used to force decompression from a given format
        even if the file has no extension, and that the default "infer" method works correctly to
        decompress a file.
        """

        # Test that manually setting a compression argument works as expected
        zip_file_path = os.path.join(THIS_DIR, 'files', 'noaa_isd_lite', 'zip_no_extension_no_missing_rows')
        analysis = diyepw.analyze_noaa_isd_lite_file(zip_file_path, compression='zip')
        self._assert_analysis_for_no_missing_rows(analysis, zip_file_path)

        # Test that the default "inferred" behavior works correctly for a format other than the
        # default GZIP that is otherwise used in all our tests and scripts
        bz2_file_path = os.path.join(THIS_DIR, 'files', 'noaa_isd_lite', 'no_missing_rows.bz2')
        analysis = diyepw.analyze_noaa_isd_lite_file(bz2_file_path)
        self._assert_analysis_for_no_missing_rows(analysis, bz2_file_path)

    def test_many_missing_rows(self):
        """
        Confirm correct behavior when too many rows are missing from a file. For this test we
        use a file that has only a single row defined.
        """
        file_path = os.path.join(THIS_DIR, 'files', 'noaa_isd_lite', 'only_one_row')
        analysis = diyepw.analyze_noaa_isd_lite_file(file_path)
        self.assertEqual(analysis['total_rows_missing'], 8759)
        self.assertEqual(analysis['max_consec_rows_missing'], 8759)

    def test_consecutive_rows(self):
        """
        Confirm that the correct number of consecutive missing rows is found in a file with sections
        of missing rows of various lengths. For this test we use a file that has lots of rows deleted,
        but the largest sequential chunk of missing rows is 17 hours long
        """
        file_path = os.path.join(THIS_DIR, 'files', 'noaa_isd_lite', 'max_17_missing_rows')
        analysis = diyepw.analyze_noaa_isd_lite_file(file_path)
        self.assertEqual(analysis['max_consec_rows_missing'], 17)

    def _assert_analysis_for_no_missing_rows(self, analysis:dict, file_name:str):
        self.assertEqual(analysis['file'], file_name)
        self.assertEqual(analysis['total_rows_missing'], 0)
        self.assertEqual(analysis['max_consec_rows_missing'], 0)


if __name__ == '__main__': # pragma: no cover
    unittest.main()



import numpy as np
import pandas as pd
import datetime
import calendar
import typing

class Meteorology:
    """
    Represents a time series of meteorological measurements at a given location.

    Attributes:
        _station_number The identification number of the weather station that made the observations
        _headers - Array - Header file data taken from an input file
        _latitude - float - Latitude at which the data was observed
        _longitude - float - Longitude at which the data was observed
        _data_source - string - Currently always "NOAA_TMY," identifies the source of the data
        _city - str - The name of the city in which the data was observed
        _state - str - The name of the city in which the data was observed
        _country - str - The name of the city in which the data was observed
        _timezone_gmt_offset - float - The timezone where the data was observed, expressed as the difference
            from GMT
        _elevation - float - The elevation at which the data was observed in meters above sea level
        _comment - str - A comment string describing the TMY file
        _observations - DataFrame - Data representing the year's observations
    """

    _RANGES = {
        'Tdb': (-70, 70),
        'Tdew': (-70, 70),
        'Patm': (31000, 120000),
        'Wspeed': (0, 40),
        'Wdir': (0, 360)
    }

    def __init__(self):
        self._data_source = 'NOAA_TMY3' # Currently hard-coded because we only process NOAA .tmy3 files

        self._station_number = None
        self._headers = None
        self._latitude = None
        self._longitude = None
        self._city = None
        self._state = None
        self._country = None
        self._timezone_gmt_offset = None
        self._elevation = None
        self._comment = None
        self._observations = None

    @property
    def station_number(self):
        return int(self._station_number)
    @station_number.setter
    def station_number(self, station_number:int):
        if station_number < 100000 or station_number > 999999:
            raise Exception("station_number must be a six-digit number")
        self._station_number = int(station_number)

    @property
    def latlong(self):
        return self._latitude, self._longitude
    @latlong.setter
    def latlong(self, latlong:typing.Tuple[float, float]):
        lat, long = latlong

        if abs(lat) > 90:
            raise Exception("Latitude must be in the range -90 - 90")
        if abs(long) > 180:
            raise Exception("Longitude must be in the range -180 - 180")

        self._latitude, self._longitude = latlong

    @property
    def city(self):
        return self._city
    @city.setter
    def city(self, city:str):
        self._city = city

    @property
    def state(self):
        return self._state
    @state.setter
    def state(self, state:str):
        self._state = state

    @property
    def country(self):
        return self._country
    @country.setter
    def country(self, country:str):
        self._country = country

    @property
    def timezone_gmt_offset(self):
        return self._timezone_gmt_offset
    @timezone_gmt_offset.setter
    def timezone_gmt_offset(self, timezone_gmt_offset:int):
        if abs(timezone_gmt_offset) > 12:
            raise Exception("timezone_gmt_offset must be in the range -12 - 12")
        self._timezone_gmt_offset = int(timezone_gmt_offset)

    @property
    def elevation(self):
        return self._elevation
    @elevation.setter
    def elevation(self, elevation:int):
        self._elevation = int(elevation)

    @property
    def observations(self):
        return self._observations
    @observations.setter
    def observations(self, observations:int):
        self._observations = observations

    def set(self, column_name:str, val):
        """
        Overwrite one of the columns in this instance's observations
        :param column_name: Must be the name of one of the columns in this instance's observations
        :param val: Mixed.
           If val is a single value, every value of the named column will be set to that value
           If val is a list or pandas.Series, it must contain the same number of items as there are
               rows in this instance's observations, and will replace those values for the named column
        :return:
        """
        if not column_name in self._observations:
            raise Exception("{c} is not one of the columns in this meteorological year's observations".format(c=column_name))

        # If we are passed a Pandas Series, just convert it to a list and let the generic list handling take care of it
        if isinstance(val, pd.Series):
            val = list(val)

        # If we have a list, we will check that its length is identical to that of the observation set, and that it has
        # consistently typed data of the correct type before using it
        if isinstance(val, list):
            # Lists replace the items in a column one-to-one, so we have to ensure that the passed list contains
            # exactly the same number of elements as our observations have rows
            if len(val) != len(self._observations):
                raise Exception("""
                    This meteorological year has {my_num} observations, but you passed a 
                    list of {your_num} replacement values
                """.format(my_num=len(self._observations), your_num=len(val))
                )

        # An assignment works here regardless of whether val is a list or a single item. If it's a list, the whole
        # column will be replaced with the values from the list. If it's a single item, that item will be duplicated
        # to fill the entire column
        self._observations[column_name] = val

    # adapted from https://github.com/SSESLab/laf/blob/master/LAF.py
    ####################################################################################################################
    # Write new EPW file
    ####################################################################################################################
    def write_epw(self, save_path):
        first_observation = {k:v[0] for (k, v) in self._observations[0:1].items()}

        location_header = ",".join([str(i) for i in [
            'LOCATION', self._city, self._state, self._country, 'customized weather file', self._station_number,
            self._latitude, self._longitude, self._timezone_gmt_offset, self._elevation
        ]])
        first_observation_date = datetime.date(
            year=int(first_observation["year"]),
            month=int(first_observation["month"]),
            day=int(first_observation["day"])
        )
        first_day_of_week = calendar.day_name[first_observation_date.weekday()]

        with open(save_path, 'w') as epw_file:
            epw_file.write("\n".join([
                location_header,
                "\n".join(self._headers[1:5]),
                self._comment,

                'COMMENTS 2, TMY3 data from energyplus.net/weather supplemented with NOAA ISD Lite data from ' +
                'https://www1.ncdc.noaa.gov/pub/data/noaa/isd-lite/ for an actual meteorological year (AMY)',

                'DATA PERIODS,1,1,Data,' + first_day_of_week + ', 1/1, 12/31'
            ]) + "\n")
            epw_file.write(self._observations.to_csv(header=False, index=False))

    # adapted from https://github.com/SSESLab/laf/blob/master/LAF.py
    @staticmethod
    def from_tmy3_file(file_path:str):
        """
        Create an instance of this class based on a tmy3 file

        :param file_path: Path to a TMY file. For the definition of a tmy3 file,
        see https://www.nrel.gov/docs/fy08osti/43156.pdf
        :return:
        """
        instance = Meteorology()

        ############################
        # Read TMY3 header
        ############################
        with open(file_path) as tmy3_file:
            instance._headers = []
            for i in range(0, 8):
                line = tmy3_file.readline().strip()
                instance._headers.append(line)

        first_line = instance._headers[0].split(',')
        instance._city = first_line[1]
        instance._state = first_line[2]
        instance._country = first_line[3]
        instance._station_number = first_line[5]
        instance._latitude = float(first_line[6])
        instance._longitude = float(first_line[7])
        instance._timezone_gmt_offset = float(first_line[8])
        instance._elevation = float(first_line[9])

        instance._comment = instance._headers[5]

        ############################
        # Read TMY3 data
        ############################
        data = np.genfromtxt(file_path, delimiter=',', skip_header=8)
        instance._observations = pd.DataFrame(data={
            "year":           [ int(i) for i in data[:, 0] ],
            "month":          [ int(i) for i in data[:, 1] ],
            "day":            [ int(i) for i in data[:, 2] ],
            "hour":           [ int(i) for i in data[:, 3] ],
            "minute":         [ int(i) for i in data[:, 4] ],
            "Flags" :         data[:, 5],
            "Tdb":            data[:, 6],
            "Tdew":           data[:, 7],
            "RH":             data[:, 8],
            "Patm":           data[:, 9],
            "ExHorRad":       data[:, 10],
            "ExDirNormRad":   data[:, 11],
            "HorIR":          data[:, 12],
            "GHRad":          data[:, 13],
            "DNRad":          data[:, 14],
            "DHRad":          data[:, 15],
            "GHIll":          data[:, 16],
            "DNIll":          data[:, 17],
            "DHIll":          data[:, 18],
            "ZenLum":         data[:, 19],
            "Wdir":           data[:, 20],
            "Wspeed":         data[:, 21],
            "TotSkyCover":    data[:, 22],
            "OpSkyCover":     data[:, 23],
            "Visib":          data[:, 24],
            "CeilH":          data[:, 25],
            "PresWeathObs":   data[:, 26],
            "PresWeathCodes": data[:, 27],
            "PrecWater":      data[:, 28],
            "AerOptDepth":    data[:, 29],
            "SnowDepth":      data[:, 30],
            "DSLS":           data[:, 31],
            "Albedo":         data[:, 32],
            "LiqPrecDepth":   data[:, 33],
            "LiqPrecQuant":   data[:, 34]
        })

        return instance

    def validate_against_epw_rules(self) -> list:
        """
        Check all observations to see whether they violate any restrictions that would prevent them
        from being used in an EPW file

        :return: A list of validation errors
        """
        violations = []

        for col_name in self._RANGES:
            min_allowed, max_allowed = self._RANGES[col_name]
            min_observed = self._observations[col_name].min()
            max_observed = self._observations[col_name].max()
            if min_observed < min_allowed or max_observed > max_allowed:
                violations.append(f"{col_name} must be in the range {min_allowed}-{max_allowed}, but this set"
                                  f" of observations includes values in the range {min_observed}-{max_observed}")

        return violations


import os
import pkg_resources
import pandas as pd

def get_wmo_station_location(wmo_station_number:int):
    """
        Get the state, and county associated with a WMO station
        :param wmo_station_number: The WMO Station ID of the weather station to get location information for
        :return: Either a dict with fields "state" and "county", or None if no information is available for
            the passed WMO Station ID
    """
    wmo_station_info_filepath = pkg_resources.resource_filename('diyepw', 'data/Weather_Stations_by_County.csv')

    if not os.path.exists(wmo_station_info_filepath):
        raise Exception("Missing WMO station info file at " + wmo_station_info_filepath)

    wmo_station_info = pd.read_csv(wmo_station_info_filepath)

    station_of_interest = wmo_station_info[wmo_station_info['Station WMO Identifier'] == wmo_station_number]

    if len(station_of_interest.index) == 0:
        return None

    station_of_interest = station_of_interest.iloc[0]

    return {
        "state": station_of_interest.State,
        "county": station_of_interest.County
    }



"""
Setup file for DIYEPW
:author:   Amanda Smith
:email:    amanda.d.smith@pnnl.gov
License:  BSD 2-Clause, see LICENSE and DISCLAIMER files
"""

import setuptools  # pragma: no coverU

def readme():
    with open('README.md') as f:
        return f.read()

setuptools.setup(  # pragma: no cover
      name='diyepw',
      version='1.0.4',
      description='A package offering support for generating EPW (EnergyPlus Weather) files',
      long_description=readme(),
      long_description_content_type="text/markdown",
      url='https://github.com/IMMM-SFA/diyepw',
      author='Amanda Smith',
      author_email='amanda.d.smith@pnnl.gov',
      packages=setuptools.find_packages(),
      package_data={ 'diyepw': ['data/**/*', 'test/files/**/*'] },
      license='BSD 2-Clause',
      python_requires='~=3.7',
      install_requires= [
          'xarray~=0.16.2',
          'numpy~=1.19.2'
      ],
      extras_require={
          'dev': [
              'pvlib~=0.8.1',
              'recommonmark~=0.7.1',
              'sphinx~=3.5.1',
              'sphinx-rtd-theme~=0.5.1'
          ]
      }
)




class DownloadNotAllowedError(Exception):
    """
    To be raised in situations where downloading a file would be necessary in order for correct
    functionality, but the caller has not given permission for file downloads to take place
    """
    pass


import pandas as pd
import os
from typing import List, Tuple
from .create_amy_epw_file import create_amy_epw_file, _tempdir_amy_epw
from ._logging import _logger

def create_amy_epw_files_for_years_and_wmos(
        years: List[int],
        wmo_indices: List[int],

        *,
        max_records_to_interpolate: int,
        max_records_to_impute: int,
        max_missing_amy_rows: int,
        amy_epw_dir: str = None,
        tmy_epw_dir: str = None,
        amy_dir: str = None,
        amy_files: Tuple[str, str] = None,
        allow_downloads: bool = False
) -> dict:
    """
    Create AMY EPW files for every combination of a set of WMO indices and years.

    Except for "years" and "wmos" being lists rather than single values, all parameters are identical in effect to
    create_amy_epw_file() - see that function's documentation for details.
    :param wmo_indices:
    :param years:
    :param max_records_to_interpolate:
    :param max_records_to_impute:
    :param max_missing_amy_rows:
    :param amy_epw_dir: Note that, in addition to the generated AMY EPW files, an errors.csv file will be created in this
        directory if any errors were encountered, with error messages explaining which year/WMO Index combinations
        failed and why.
    :param tmy_epw_dir:
    :param amy_dir:
    :param amy_files:
    :param allow_downloads:
    :return: A dictionary of the files generated for each year/wmo combination, in the form {
        <year>: {
            <wmo_index>: [<file_path>, ...],
            ...
        },
        ...
    }
    """
    amy_epw_files = {}

    # Initialize the df to hold information about year/WMO Index combinations for which no EPW could be generated
    errors = pd.DataFrame(columns=['year', 'wmo_index', 'error'])
    errors_path = os.path.join(amy_epw_dir if amy_epw_dir else _tempdir_amy_epw, "errors.csv")

    for year in years:
        for wmo_index in wmo_indices:
            _logger.info(f"Creating AMY EPW for year {year} and WMO {wmo_index}...")

            if not year in amy_epw_files:
                amy_epw_files[year] = {}
            if not wmo_index in amy_epw_files[year]:
                amy_epw_files[year][wmo_index] = []

            try:
                amy_epw_file = create_amy_epw_file(
                    wmo_index,
                    year,
                    max_records_to_interpolate=max_records_to_interpolate,
                    max_records_to_impute=max_records_to_impute,
                    max_missing_amy_rows=max_missing_amy_rows,
                    amy_epw_dir=amy_epw_dir,
                    tmy_epw_dir=tmy_epw_dir,
                    amy_dir=amy_dir,
                    amy_files=amy_files,
                    allow_downloads=allow_downloads
                )
                amy_epw_files[year][wmo_index].append(amy_epw_file)
            except Exception as e:
                errors = errors.append({"year": year, "wmo_index": wmo_index, "error": str(e)}, ignore_index=True)
                print(f'Problem processing year {year} and WMO index {wmo_index}: ' + str(e))

    if not errors.empty:
        _logger.warning(f"AMY EPW files could not be generated for {len(errors)} year/WMO Index combinations - see {errors_path} for more information")
        errors.to_csv(errors_path, mode='w', index=False)

    return amy_epw_files
